"use strict";

var Immutable = require('immutable');
var isIterable = Immutable.Iterable.isIterable;


function makeKeyedSeqByField(data, keyField) {
  return Immutable.Map().withMutations(function (map) {
    data.forEach(function (datum) {
      var value = datum.get(keyField);

      if (value === undefined) {
        throw new Error('Every indexed document must have a key field. Currently looking ' + ('for keys in the property "' + keyField + '" which is not present in ') + ('document:\n\n' + JSON.stringify(datum) + '\n'));
      }

      if (map.has(value)) {
        throw new Error('Multiple documents have identical key value ' + ('(currently set to "' + keyField + '"): ' + value));
      }

      map.set(value, datum);
    });
  });
}

var makeSelection = function makeSelection(name, values) {
  return Immutable.Map({
    name: name,
    values: (isIterable(values) ? values : Immutable.fromJS(values)).toSet()
  });
};

var withoutName = function withoutName(set, name) {
  return set.filter(function (item) {
    return item.get('name') !== name;
  });
};

var FacetSet = Immutable.Record({
  data: null,
  keyField: null,
  facets: Immutable.Map(),
  selections: Immutable.OrderedSet(),
  _selectionResultCache: Immutable.Map()
});

/* These will return new FacetSet objects */
Object.assign(FacetSet.prototype, {
  addFacet: function addFacet(name, fn, opts) {
    var _this = this;

    return this.delete('_selectionResultCache').update('facets', function (facets) {
      return facets.set(name, makeFacet(_this, fn, opts));
    });
  },
  addFieldFacet: function addFieldFacet(field, opts) {
    var path = Array.isArray(field) ? field : field.split('.'),
        accessor = function accessor(d) {
      return d.getIn(path);
    };

    return this.addFacet(path.join('.'), accessor, opts);
  },
  removeFacet: function removeFacet(name) {
    return this.delete('_selectionResultCache').update('facets', function (facets) {
      return facets.delete(name);
    }).update('selections', function (selections) {
      return withoutName(selections, name);
    });
  },
  addSelection: function addSelection(name, values) {
    if (!this.facets.has(name)) {
      throw new Error('No such facet: ' + name);
    }

    return this.update('selections', function (selections) {
      return selections.add(makeSelection(name, values));
    });
  },
  removeSelection: function removeSelection(name, values) {
    return this.update('selections', function (selections) {
      return selections.delete(makeSelection(name, values));
    });
  },
  resetSelections: function resetSelections(name) {
    return this.update('selections', function (selections) {
      return name ? withoutName(selections, name) : selections.clear();
    });
  },
  facetsAfterSelections: function facetsAfterSelections() {
    var _this2 = this;

    var opts = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];
    var forIDs = opts.forIDs;
    var forFields = opts.forFields;
    var fields = forFields || this.facets.keySeq();

    var selectedIDs = this.selectedIDs();

    return Immutable.Map().withMutations(function (map) {
      Immutable.List(fields).forEach(function (name) {
        var facet = _this2.facets.get(name);

        var matchIDs = void 0;

        if (_this2.selections.size) {
          matchIDs = selectedIDs;
        }

        if (forIDs) {
          matchIDs = matchIDs ? matchIDs.intersect(forIDs) : Immutable.Set(forIDs);
        }

        map.set(name, !matchIDs ? facet : facet.map(function (ids) {
          return ids.intersect(matchIDs);
        }).filter(function (ids) {
          return ids.size;
        }));
      });
    });
  },
  facetValuesAfterSelections: function facetValuesAfterSelections() {
    var opts = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];

    return this.facetsAfterSelections(opts).map(function (field) {
      return field.keySeq().toSet();
    });
  },
  selectedIDs: function selectedIDs() {
    var _this3 = this;

    return this.selections.size === 0 ? this.data.keySeq() : this.selections.map(function (selection) {
      return matchedIDsForSelection(_this3, selection);
    }).reduce(function (prev, next) {
      return prev.intersect(next);
    });
  },
  selectedItems: function selectedItems() {
    var _this4 = this;

    return this.selectedIDs().map(function (id) {
      return _this4.data.get(id);
    });
  }
});

function makeFacet(facetSet, fn) {
  var opts = arguments.length <= 2 || arguments[2] === undefined ? {} : arguments[2];
  var multiValue = opts.multiValue;
  var facet = Immutable.Map().asMutable();

  facetSet.data.forEach(function (item) {
    var key = item.get(facetSet.keyField);

    var result = fn(item);

    result = isIterable(result) ? result : Immutable.fromJS(result);

    var values = multiValue ? Immutable.List(result) : Immutable.List.of(result);

    values.forEach(function (value) {
      facet.update(value, Immutable.Set(), function (set) {
        return set.asMutable().add(key);
      });
    });
  });

  return facet.map(function (set) {
    return set.asImmutable();
  }).asImmutable();
}

function matchedIDsForSelection(facetSet, selection) {
  var name = selection.get('name'),
      selectedValues = selection.get('values'),
      facet = facetSet.facets.get(name);

  return facet.filter(function (itemIDs, facetValue) {
    return selectedValues.contains(facetValue);
  }).toSet().flatten();
}

module.exports = function (data) {
  var keyField = arguments.length <= 1 || arguments[1] === undefined ? 'id' : arguments[1];

  return new FacetSet({
    data: makeKeyedSeqByField(data, keyField),
    keyField: keyField
  });
};